from urllib.parse import quote


def _dict_to_qs(params):
    """Converts a dict to a query string, with values url-encoded"""
    qs = []
    for k, v in params.items():
        qs.append(f"{quote(k)}={quote(str(v))}")
    return "&".join(qs)


def make_issue_link(*, title: str, body: str, labels: list[str] = []) -> str:
    """Makes a generic GH issue"""
    url = "https://github.com/amoffat/HeimdaLLM/issues/new"
    params = {
        "title": title,
        "body": body,
        "labels": ",".join(labels),
    }
    return f"{url}?{_dict_to_qs(params)}"


def make_ambiguous_parse_issue(query, trees):
    """Makes a GH issue for an ambiguous parse"""

    def fmt_tree(i, tree):
        return f"""
## Tree {i}
```\n{tree.pretty()}```
""".strip()

    nice_trees = "\n\n".join(fmt_tree(i, tree) for i, tree in enumerate(trees))
    body = f"""# Query

```sql
{query}
```

# Parse trees
{nice_trees}
""".strip()
    return make_issue_link(
        title="Ambiguous parse",
        body=body,
        labels=["bug", "ambiguous parse"],
    )



from heimdallm.bifrosts.sql.validator import (
    ConstraintValidator as _SQLConstraintValidator,
)

from .. import presets


class ConstraintValidator(_SQLConstraintValidator):
    __doc__ = _SQLConstraintValidator.__doc__

    def can_use_function(self, function: str) -> bool:
        """
        Returns whether or not a SQL function is allowed to be used anywhere in the
        query. By default, this checks the function against the list of safe functions
        that we have curated by hand.

        :param function: The *lowercase* name of the function.
        :return: Whether or not the function is allowed.
        """
        return function in presets.safe_functions



import re
from abc import abstractmethod
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Sequence, cast

import jinja2

from heimdallm.envelope import PromptEnvelope as _BasePromptEnvelope
from heimdallm.llm import LLMIntegration

if TYPE_CHECKING:
    import heimdallm.bifrosts.sql.validator

THIS_DIR = Path(__file__).parent
_TMPL_ENV = jinja2.Environment(
    loader=jinja2.FileSystemLoader(THIS_DIR / "envelopes"),
    undefined=jinja2.StrictUndefined,
)


class PromptEnvelope(_BasePromptEnvelope):
    """The purpose of the prompt envelope is to wrap the untrusted input in additional
    context for the LLM to produce the correct output. We do not do validation in the
    envelope, because it is impossible to prevent prompt injection.

    While not necessary to subclass, you are recommended to do so if you want to
    customize the envelope.

    :param llm: The LLM integration being sent the human input. This can be used to
        tweak the :meth:`wrap` and :meth:`unwrap` methods to account for quirks of the
        specific LLM.
    :param db_schema: The database schema of the database being queried. It is passed to
        the LLM so that the LLM knows how the tables and columns are connected.
    :param validators: The validators to use to validate the output of the LLM. They
        aren't used to validate here, but some of the validator's properties are added
        to the envelope to help guide the LLM to produce the correct output.
    """

    def __init__(
        self,
        *,
        llm: LLMIntegration,
        db_schema: str,
        validators: Sequence["heimdallm.bifrosts.sql.validator.ConstraintValidator"],
    ):
        self.db_schema = db_schema
        self.validators = validators
        super().__init__(llm=llm)

    @abstractmethod
    def template(self, env: jinja2.Environment) -> jinja2.Template:
        """
        Returns the template to use for the envelope. Override in a subclass for
        complete customization.

        :param env: The environment to use to load the template.
        :return: The template to use for the envelope.
        """
        raise NotImplementedError

    @property
    def params(self) -> dict:
        """Returns a dictionary of additional parameters to be passed to the template.
        Override in a subclass for complete control over values that you want in the
        envelope.

        :return: The extra parameters to pass to the template."""
        return {}

    def wrap(self, untrusted_input: str) -> str:
        """Performs the wrapping of the untrusted input with the envelope. Not
        intended to be overridden, but not foribben either. Consider overriding the
        :meth:`template` and :meth:`params` properties first instead.

        :param untrusted_input: The untrusted input from the user.
        :return: The wrapped input to send to the LLM."""
        all_idents = chain.from_iterable(
            v.requester_identities() for v in self.validators
        )
        id_constraints = " or ".join(str(ident) for ident in all_idents)
        params = {
            "id_constraints": id_constraints,
            "db_schema": self.db_schema,
            "query": untrusted_input,
        }
        params.update(self.params)

        tmpl = self.template(_TMPL_ENV)
        prompt = tmpl.render(params)
        return prompt

    def unwrap(self, untrusted_llm_output: str) -> str:
        """Unpack the SQL query from the LLM output by finding it (hopefully) among the
        delimiters.

        :param untrusted_llm_output: The output from the LLM.
        :return: The SQL query."""
        # assume the LLM did what we said and delimited the output with ```
        if "```" in untrusted_llm_output:
            # sometimes the LLM is silly and likes to include the word "sql" inside
            # the delimiters. silly LLM!
            match = cast(
                re.Match,
                re.search(
                    r"```(?:sql)?(.*?)```",
                    untrusted_llm_output,
                    flags=re.DOTALL | re.IGNORECASE,
                ),
            )
            unpacked = match.group(1)
        # otherwise, just return the whole thing, and we'll attempt to parse it
        # as is
        else:
            # for some reason, the LLM is returning "sql\n" as a prefix, even
            # when instructed not to, oh well.
            unpacked = re.sub(r"^\s*sql\n+", "", untrusted_llm_output)

        unpacked = unpacked.strip()
        return unpacked


class TestSQLPromptEnvelope(PromptEnvelope):
    """
    A test envelope specifically for :meth:`Bifrost.mocked
    <heimdallm.bifrosts.sql.bifrost.Bifrost.mocked>`

    :param llm: The LLM integration being sent the human input. This can be used to
        tweak the :meth:`wrap` and :meth:`unwrap() <PromptEnvelope.unwrap>` methods to
        account for quirks of the specific LLM.
    :param db_schema: The database schema of the database being queried. It is passed to
        the LLM so that the LLM knows how the tables and columns are connected.
    :param validators: The validators to use to validate the output of the LLM. They
        aren't used to validate here, but some of the validator's properties are added
        to the envelope to help guide the LLM to produce the correct output.

    :meta private:
    """

    def wrap(self, untrusted_input: str) -> str:
        # build the envelope, but don't actually use it. this exercises the
        # envelope code as part of a full integration
        super().wrap(untrusted_input)
        return untrusted_input

    def template(self, env: jinja2.Environment) -> jinja2.Template:
        """
        Returns the template to use for the envelope. Override in a subclass for
        complete customization.

        :param env: The environment to use to load the template.
        :return: The template to use for the envelope.
        """
        return env.get_template("sql/test.j2")



from abc import ABC, abstractmethod

from heimdallm.llm import LLMIntegration


class PromptEnvelope(ABC):
    """
    The purpose of the prompt envelope is to guide the wrapped untrusted query to
    produce the structured output. It is impossible to prevent prompt injection, so
    accept that up front. However, a malicious prompt will not produce a valid response,
    so our grammar will not parse it, and the validator will not validate it, so the
    malicious user will never see the result of their malicious output.

    :param llm: The LLM integration being sent the human input. This is passed in so
        that the envelope can change the way it wraps or unwraps data communicated with
        the LLM, based on different quirks of the LLM.
    """

    def __init__(self, *, llm: LLMIntegration):
        self.llm = llm

    @abstractmethod
    def wrap(self, untrusted_input: str) -> str:
        """Wrap the untrusted input with additional context to guide the LLM to produce
        the correct output.

        :param untrusted_input: The untrusted input from the user.
        :return: The wrapped input to send to the LLM."""
        raise NotImplementedError

    @abstractmethod
    def unwrap(self, untrusted_llm_output: str) -> str:
        """Unwrap the LLM's output to produce the original untrusted input. This method
        should work closely with the wrap method to coordinate how to delimit the
        structured response.

        :param untrusted_llm_output: The untrusted output from the LLM.
        :return: The structured response, unwrapped from the LLM's output."""
        raise NotImplementedError



from lark import Token, Tree, Visitor

from ..utils.identifier import get_identifier


class AliasCollector(Visitor):
    """collects all of our table and column aliases and maps them back to the
    authoritative name. we have to do this pass first, before any other passes,
    because the tree may not evaluate in the order that would allow us to
    resolve aliases."""

    def __init__(self, reserved_keywords: set[str]):
        # aliased table names from the FROM clause, as well as JOIN clauses.
        # they map from the alias name to the table name.
        self._aliased_tables: dict[str, str] = {}
        # aliased column names from the SELECT clause. they map from the alias
        # name to the (table, column) tuple
        self._aliased_columns: dict[str, tuple[str | None, str | None]] = {}
        self._reserved_keywords = reserved_keywords

    def _resolve_table(self, table):
        return self._aliased_tables.get(table, table)

    # tables are aliased in the FROM clause of a SELECT statement, or when a
    # table is JOINed. it's here that we know the authoritative table name and
    # its aliased, which may be used in other parts of the query
    def aliased_table(self, node: Tree):
        table_node, _as, alias_node = node.children
        table_name = get_identifier(table_node, self._reserved_keywords)
        alias = get_identifier(alias_node, self._reserved_keywords)
        self._aliased_tables[alias] = table_name

        # inefficient, but backfill the correct table alias for any columns
        for key, (tn, cn) in list(self._aliased_columns.items()):
            if tn == alias:
                self._aliased_columns[key] = (table_name, cn)

    # columns are aliased in the column list of a SELECT statement. we assume
    # that all aliased columns are fully qualified by table name. note, however,
    # that the table name may be an alias itself.
    def aliased_column(self, node: Tree):
        value_node, _as, alias_node = node.children
        alias_name = get_identifier(alias_node, self._reserved_keywords)

        if isinstance(value_node, Token):
            if value_node.type == "COUNT_STAR":
                self._aliased_columns[alias_name] = (None, None)
            return

        # there may be multiple columns referenced in this alias column, for
        # example if we're running a function on multiple columns and aliasing
        # the result. so we need to look at each fq_column individually
        for fq_column_node in value_node.find_data("fq_column"):
            table_node, column_node = fq_column_node.children
            table_name = get_identifier(table_node, self._reserved_keywords)
            column_name = get_identifier(column_node, self._reserved_keywords)

            # do we already have a table alias for this column? use that instead for
            # the table name
            table_name = self._resolve_table(table_name)
            self._aliased_columns[alias_name] = (table_name, column_name)



from typing import TYPE_CHECKING, Callable, Sequence

import structlog
from lark import Lark, ParseTree

if TYPE_CHECKING:
    import heimdallm.constraints
    import heimdallm.envelope
    from heimdallm.llm import LLMIntegration

LOG = structlog.get_logger(__name__)


class Bifrost:
    """The Bifrost is the bridge from the outside world to your secure systems. It is
    responsible for a rigorous parsing and validation of the output of the :term:`LLM`.
    This class is not used directly, but is subclassed to provide a Bifrost for a
    specific problem, like constructing trusted SQL queries. It is general enough to be
    adapted to other problems.

    :param llm: The LLM integration to use.
    :param prompt_envelope: The prompt envelope used to wrap the untrusted human input
        and unwrap the untrusted LLM output.
    :param grammar: The grammar that defines the structured output that we expect from
        the LLM.
    :param tree_producer: A callable that takes a grammar and the untrusted input and
        returns a parse tree. Using a callback allows us to do Bifrost-specific parsing,
        for example, to collapse ambiguous parse trees into a single parse tree.
    :param constraint_validators: A sequence of constraint validators that will be used
        to validate the parse tree returned by the ``tree_producer``. Only one validator
        needs to succeed for validation to pass.
    """

    def __init__(
        self,
        *,
        llm: "LLMIntegration",
        prompt_envelope: "heimdallm.envelope.PromptEnvelope",
        grammar: Lark,
        tree_producer: Callable[[Lark, str], ParseTree],
        constraint_validators: Sequence["heimdallm.constraints.ConstraintValidator"],
    ):
        self.llm = llm
        self.prompt_envelope = prompt_envelope
        self.grammar = grammar
        self.tree_producer = tree_producer
        self.constraint_validators = constraint_validators

    def traverse(
        self,
        untrusted_human_input: str,
        autofix: bool = True,
    ) -> str:
        """Run the full chain of the Bifrost, from untrusted input to trusted
        input. Traversing the Bifrost means successfully returning a value from this
        function, which is only possible if every step succeeds.

        This is the main entry point for the Bifrost API.

        :param untrusted_human_input: The untrusted input from the user.
        :param autofix: Whether or not to attempt to :doc:`reconstruct
            </reconstruction>` the input to satisfy the constraint validator.

        :return: The trusted LLM output.
        """

        log = LOG.bind(input=untrusted_human_input, autofix=autofix)
        log.info("Traversing untrusted input")

        # wrap the untrusted input in our prompt
        log.info("Wrapping input in prompt envelope")
        untrusted_llm_input = self.prompt_envelope.wrap(untrusted_human_input)
        log.debug("Produced prompt envelope")

        # talk to our LLM
        log.info("Sending envelope to LLM")
        untrusted_llm_output: str = self.llm.complete(untrusted_llm_input)
        log = log.bind(llm_output=untrusted_llm_output)
        log.info("Received raw result from LLM")

        # trim any cruft off of the LLM output
        log.info("Unwrapping prompt envelope")
        try:
            untrusted_llm_output = self.prompt_envelope.unwrap(untrusted_llm_output)
        except Exception as e:
            log.exception("Unwrap failed")
            raise e
        log = log.bind(unwrapped=untrusted_llm_output)
        log.info("Unwrap succeeded")

        # throws a parse error
        log.info("Parsing result via grammar")
        try:
            tree = self.parse(untrusted_llm_output)
        except Exception as e:
            log.exception("Parse failed")
            raise e
        log.info("Parse succeeded")

        # try each of our constraint validators until one succeeds
        validation_exc = None
        for validator in self.constraint_validators:
            log.info(
                "Trying constraint validator",
                validator=validator.__class__.__name__,
            )
            try:
                trusted_llm_output = self._try_validator(
                    log,
                    validator,
                    autofix,
                    untrusted_llm_output,
                    tree,
                )
            except Exception as e:
                validation_exc = e
            # the first validator to validate the input wins
            else:
                validation_exc = None
                break

        if validation_exc:
            # ugly, but this is the easiest way to log the exception that we have
            # already caught
            try:
                raise validation_exc
            except Exception as e:
                log.exception("Validation failed")
                raise e

        log = log.bind(trusted=untrusted_llm_output)
        log.info("Validation succeeded")

        return trusted_llm_output

    def _try_validator(
        self,
        log: structlog.BoundLogger,
        validator: "heimdallm.constraints.ConstraintValidator",
        autofix: bool,
        untrusted_llm_output: str,
        tree: ParseTree,
    ) -> str:
        """Attempt validation with an individual constraint validator."""

        if autofix:
            log.info("Autofixing parse tree and reconstructing the input")
            try:
                untrusted_llm_output = validator.fix(self, self.grammar, tree)
            except Exception as e:
                log.exception("Autofix failed")
                raise e
            log.info("Re-parsing reconstructed output")
            try:
                tree = self.parse(untrusted_llm_output)
            except Exception as e:
                log.exception("Reparse failed")
                raise e
            log.info("Reconstruction succeeded")

        # throws a bifrost-specific exception
        log.info("Validating parse tree")
        validator.validate(self, untrusted_llm_output, tree)
        log.info("Validation succeeded")

        return untrusted_llm_output

    def parse(self, untrusted_llm_output: str) -> ParseTree:
        """Converts the :term:`LLM` output into a parse tree. Override it in a subclass
        to throw custom exceptions based on the grammar and parse state.

        :param untrusted_llm_output: The unwrapped output from the LLM.

        :return: The parse tree.
        """
        return self.tree_producer(self.grammar, untrusted_llm_output)



from pathlib import Path

from lark import Lark

from heimdallm.bifrosts.sql.bifrost import Bifrost as _SQLBifrost

from .. import presets

_THIS_DIR = Path(__file__).parent
_GRAMMAR_PATH = _THIS_DIR / "grammar.lark"


class Bifrost(_SQLBifrost):
    """
    A Bifrost for MySQL ``SELECT`` queries

    :param llm: The LLM integration to use.
    :param prompt_envelope: The prompt envelope used to wrap the untrusted human input
        and unwrap the untrusted LLM output.
    :param constraint_validators: A sequence of constraint validators that will be used
        to validate the parse tree returned by the ``tree_producer``. Only one validator
        needs to succeed for validation to pass.
    """

    @staticmethod
    def build_grammar() -> Lark:
        """
        Returns a limited ``SELECT`` grammar

        noteworthy:
            - no outer joins
            - no subqueries

        Theoretically, subqueries could be allowed, but it would be more work, and
        I'm not yet convinced that an LLM produces subqueries often enough to make
        it worth it.

        Outer joins are unsafe because the join constraint is not applied to the
        rows that would be considered "outer."
        """
        with open(_GRAMMAR_PATH, "r") as h:
            grammar = Lark(
                ambiguity="explicit",
                maybe_placeholders=False,
                grammar=h,
            )
        return grammar

    @classmethod
    def reserved_keywords(self) -> set[str]:
        return presets.reserved_keywords
